unit FavSemaphore;

// ****************************************************
// Compiler: CodeTyphon 6.00
// ****************************************************

// ****************************************************
// Used for:
// This is mutex and semaphore objects.
// ****************************************************

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,
  {$ifdef Windows} Windows {$endif}
  {$ifdef Linux}   Unix, unixtype, pthreads, BaseUnix {$endif};

{$ifdef Linux}
const
   EPERM = 1;
   ENOENT = 2;
   ESRCH = 3;
   EINTR = 4;
   EIO = 5;
   ENXIO = 6;
   E2BIG = 7;
   ENOEXEC = 8;
   EBADF = 9;
   ECHILD = 10;
   EAGAIN = 11;
   ENOMEM = 12;
   EACCES = 13;
   EFAULT = 14;
   ENOTBLK = 15;
   EBUSY = 16;
   EEXIST = 17;
   EXDEV = 18;
   ENODEV = 19;
   ENOTDIR = 20;
   EISDIR = 21;
   EINVAL = 22;
   ENFILE = 23;
   EMFILE = 24;
   ENOTTY = 25;
   ETXTBSY = 26;
   EFBIG = 27;
   ENOSPC = 28;
   ESPIPE = 29;
   EROFS = 30;
   EMLINK = 31;
   EPIPE = 32;
   EDOM = 33;
   ERANGE = 34;
   EDEADLK = 35;
   ENAMETOOLONG = 36;
   ENOLCK = 37;
   ENOSYS = 38;
   ENOTEMPTY = 39;
   ELOOP = 40;
   EWOULDBLOCK = EAGAIN;
   ENOMSG = 42;
   EIDRM = 43;
   ECHRNG = 44;
   EL2NSYNC = 45;
   EL3HLT = 46;
   EL3RST = 47;
   ELNRNG = 48;
   EUNATCH = 49;
   ENOCSI = 50;
   EL2HLT = 51;
   EBADE = 52;
   EBADR = 53;
   EXFULL = 54;
   ENOANO = 55;
   EBADRQC = 56;
   EBADSLT = 57;
   EDEADLOCK = EDEADLK;
   EBFONT = 59;
   ENOSTR = 60;
   ENODATA = 61;
   ETIME = 62;
   ENOSR = 63;
   ENONET = 64;
   ENOPKG = 65;
   EREMOTE = 66;
   ENOLINK = 67;
   EADV = 68;
   ESRMNT = 69;
   ECOMM = 70;
   EPROTO = 71;
   EMULTIHOP = 72;
   EDOTDOT = 73;
   EBADMSG = 74;
   EOVERFLOW = 75;
   ENOTUNIQ = 76;
   EBADFD = 77;
   EREMCHG = 78;
   ELIBACC = 79;
   ELIBBAD = 80;
   ELIBSCN = 81;
   ELIBMAX = 82;
   ELIBEXEC = 83;
   EILSEQ = 84;
   ERESTART = 85;
   ESTRPIPE = 86;
   EUSERS = 87;
   ENOTSOCK = 88;
   EDESTADDRREQ = 89;
   EMSGSIZE = 90;
   EPROTOTYPE = 91;
   ENOPROTOOPT = 92;
   EPROTONOSUPPORT = 93;
   ESOCKTNOSUPPORT = 94;
   EOPNOTSUPP = 95;
   EPFNOSUPPORT = 96;
   EAFNOSUPPORT = 97;
   EADDRINUSE = 98;
   EADDRNOTAVAIL = 99;
   ENETDOWN = 100;
   ENETUNREACH = 101;
   ENETRESET = 102;
   ECONNABORTED = 103;
   ECONNRESET = 104;
   ENOBUFS = 105;
   EISCONN = 106;
   ENOTCONN = 107;
   ESHUTDOWN = 108;
   ETOOMANYREFS = 109;
   ETIMEDOUT = 110;
   ECONNREFUSED = 111;
   EHOSTDOWN = 112;
   EHOSTUNREACH = 113;
   EALREADY = 114;
   EINPROGRESS = 115;
   ESTALE = 116;
   EUCLEAN = 117;
   ENOTNAM = 118;
   ENAVAIL = 119;
   EISNAM = 120;
   EREMOTEIO = 121;
   EDQUOT = 122;
   ENOMEDIUM = 123;
   EMEDIUMTYPE = 124;

   (*
   LOCK_SH = 1;
   LOCK_EX = 2;
   LOCK_NB = 4;
   LOCK_UN = 8;
   *)

   SEM_FAILED    = Psem_t(nil);
   SEM_VALUE_MAX = ((not 0) shr 1);
{$endif}

const
   FAV_SEM_CREATE = 0;
   FAV_SEM_READ_LOCK = 1;
   FAV_SEM_READ_UNLOCK = 2;
   FAV_SEM_WRITE_LOCK = 3;
   FAV_SEM_WRITE_UNLOCK = 4;
   FAV_SEM_DESTROY = 5;

type

  TFavRecSyncObjectsStatus = record
    Name: String;                // name of mutex, semaphore....
    ThreadID: Int64;             // thread id
    CreateCnt: Int64;            // create area
    CreateTime: TDateTime;
    CreateLocation: String;      // Source code Location, programmer writes location

    ReadCnt: Int64;              // read area
    ReadTime: TDateTime;
    ReadLocation: String;
    ReadUnlockCnt: Int64;              // read area
    ReadUnlockTime: TDateTime;
    ReadUnlockLocation: String;

    WriteCnt: Int64;             // write area
    WriteTime: TDateTime;
    WriteLocation: String;
    WriteUnlockCnt: Int64;             // write area
    WriteUnlockTime: TDateTime;
    WriteUnlockLocation: String;

    DestroyCnt: Int64;           // destroy area
    DestroyTime: TDateTime;
    DestroyLocation: String;
  end;

  { TFavSwitch }

  TFavSwitch = class
  private
    _Value: Boolean;
  public
    constructor Create();
    destructor  Destroy; override;
  published
    property Value: Boolean read _Value write _Value;
  end;

  { TFavMutexDebug }

  TFavMutexDebug = class
  private
    _LastErroNo: longint;
    _LastErrorMsg: String;
    _Name: String;
    _InitialValue: Boolean;
    {$ifdef Windows}
    FMutex: Cardinal;
    {$endif}
    {$ifdef Linux}
    FMutex: pthread_mutex_t;
    FMutexFile: Longint;
    {$endif}
  public
    FObjectName: String;
    constructor Create(InObjectName: String; InLocation: String; InName: String = ''; InInitialValue: Boolean=False);
    destructor  Destroy; override;
    procedure   Release(InLocation: String);
    procedure   Wait(InLocation: String);
    function GetError(InAction: String; InError: Int64): String;
  end;

  { TFavMutex }

  TFavMutex = class
  private
    _LastErroNo: longint;
    _LastErrorMsg: String;
    _Name: String;
    _InitialValue: Boolean;
    {$ifdef Windows}
    FMutex: Cardinal;
    {$endif}
    {$ifdef Linux}
    FMutex: pthread_mutex_t;
    FMutexFile: Longint;
    {$endif}
  public
    constructor Create(InName: String = ''; InInitialValue: Boolean=False);
    destructor  Destroy; override;
    procedure   Release;
    procedure   Wait;
    function GetError(InAction: String; InError: Int64): String;
  end;

  { TFavSyncObjectsStatus }

  TFavSyncObjectsStatus = class
  public
    _data: TFavRecSyncObjectsStatus;
    constructor Create;
    destructor  Destroy; override;
  end;

  { TFavSyncObjectsStatuses }

  TFavSyncObjectsStatuses = class
  public
    _Controller: TFavMutex;
    _List: TList;
    constructor Create;
    destructor  Destroy; override;
    procedure RegisterAction(InAction: Integer; InThreadID: Int64; InName: String; InLocation: String);
    procedure DeleteAllItems;
    function FindItem(InName: String; InThreadID: Int64): TFavSyncObjectsStatus;
    function  DateTimeToString(InDateTime: TDateTime): String;
    function  GenerateDumpList: String;
    procedure SaveToFile(InFileName: String; InString: String);
    procedure AppendToFile(InFileName: String; InString: String);
    procedure CleanSynchronizedLines;
    procedure DumpToFile(InCleanList: Boolean);
  end;

  { TFavSemaphore }

  TFavSemaphore = class
  private
    _LastErroNo: Longint;
    _LastErrorMsg: String;
    _Name: String;
    _InitialValue: Longint;
    _MaxValue: Longint;
    {$ifdef Windows}
    FSemaphore: Cardinal;
    {$endif}
    {$ifdef Linux}
    FSemaphore: sem_t;
    PSemaphore: Psem_t;
    {$endif}
  public
    constructor Create(InName: String = ''; InInitialValue: Cardinal = 1; InMaxValue: Cardinal = 1);
    destructor  Destroy; override;
    procedure   Release; overload;
    procedure   Release(HowMuch: Integer); overload;
    procedure   Wait;
    function GetError(InAction: String; InError: Int64): String;
  end;

  { TFavWriterReaderSemaphoreSimple }

  TFavWriterReaderSemaphoreSimple = class
    public
      FObjectName: String;
      _mutexMain: TFavMutex;
      constructor Create(InObjectName: String; InLocation: String);
      destructor  Destroy; override;

      procedure WaitForRead(InLocation: String);
      procedure WaitForWrite(InLocation: String);
      procedure ReleaseRead(InLocation: String);
      procedure ReleaseWrite(InLocation: String);
  end;


  { TFavWriterReaderSemaphoreWriterPriorityFPC }

  // the problem with this implementation is that this
  // implementation behave almost like mutex esspecially
  // when write action is much bigger then read action - for example 50 times
  // Also it is not good when you have much writters - only one is allowed
  TFavWriterReaderSemaphoreWriterPriorityFPC = class
    _sync: TMultiReadExclusiveWriteSynchronizer;
    public
      FObjectName: String;
      constructor Create(InObjectName: String; InLocation: String);
      destructor  Destroy; override;

      procedure WaitForRead(InLocation: String);
      procedure WaitForWrite(InLocation: String);
      procedure ReleaseRead(InLocation: String);
      procedure ReleaseWrite(InLocation: String);
  end;

  { TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi }

  TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi = class
    private
      FActiveReaders  : integer;
      FActiveWriter   : boolean;
      FBlockedReaders : TFavSemaphore;
      FBlockedWriters : TFavSemaphore;
      FController     : TFavMutex;
      FWaitingReaders : integer;
      FWaitingWriters : integer;
      FMaxReaders     : Integer;
    protected
    public
      FObjectName: String;
      constructor Create(InObjectName: String; InLocation: String; InMaxReaders: Integer = 4096 * 4096);
      // constructor Create(InMaxReaders: Integer = 2);
      destructor Destroy; override;

      procedure WaitForRead(InLocation: String);
      procedure WaitForWrite(InLocation: String);
      procedure ReleaseRead(InLocation: String);
      procedure ReleaseWrite(InLocation: String);
  end;






  //TFavWriterReaderSemaphoreWriterPriority = TFavWriterReaderSemaphoreSimple;
  //TFavWriterReaderSemaphoreWriterPriority = TFavWriterReaderSemaphoreWriterPriorityFPC;
  TFavWriterReaderSemaphoreWriterPriority = TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi;












  { TFavQueueData }

  TFavQueueData = class
  public
    _data: String;
    constructor Create(InData: String = '');
    destructor  Destroy; override;
  end;

  TFavQueue = class
    _maxMsgNo: Longint;
    _list: TList;
    _mutexAccess: TFavMutex;
    _semaphoreReaders: TFavSemaphore;
    _semaphoreWriters: TFavSemaphore;
    constructor Create(InMaxMsgNo: Longint);
    destructor  Destroy; override;
    procedure PutData(var InString: String);
    function GetData: String;
    procedure CleanQueue;
  end;

  { TThreadSyncObjectStatus }

  TThreadSyncObjectStatus = class(TThread)
  private
  protected
    procedure Execute; override;
  public
    _BShutdown, _BTerminated: Boolean;
    _Counter: Int64;
    constructor Create(CreateSuspended: Boolean;
                       const StackSize: SizeUInt = DefaultStackSize);
    destructor Destroy; override;
    procedure ShutDownThread;
  end;


{$ifdef Linux}
//function sem_init(__sem:Psem_t; __pshared:cint; __value:dword):cint;cdecl;external libthreads;
function sem_open(__name:Pchar; __oflag:longint; args:array of const):Psem_t;cdecl;external libthreads;
function sem_open(__name:Pchar; __oflag:longint):Psem_t;cdecl;varargs;external libthreads;
{$endif}

function FavGenerateUniqueSyncObjectName(InPrefix: String): String;

const InitTFavRecSyncObjectStatus: TFavRecSyncObjectsStatus = ({%H-});

var FAV_SYNCOBJECT_STATUS: TFavSyncObjectsStatuses;
    FAV_THREAD_SYNCOBJECT: TThreadSyncObjectStatus;
    FAV_THREAD_SYSOBJECTS_CLEAN_TIMEOUT: Integer;      // timeout in seconds when thread will call cleanup
    FAV_RAISE_SYNCOBJECTS_REGISTRATION_ERROR: Boolean; // if true program will raise exception when it gets registration error

implementation

function FavGenerateUniqueSyncObjectName(InPrefix: String): String;
begin
  Result := InPrefix + '_' + IntToStr(GetTickCount) + '_' + IntToStr(GetThreadID);
end;


{ TThreadSyncObjectStatus }

procedure TThreadSyncObjectStatus.Execute;
var MyCount: Integer;
begin
  _BShutdown := False;
  _BTerminated := False;
  MyCount := 0;
  while true do begin
    Inc(MyCount);
    if MyCount = FAV_THREAD_SYSOBJECTS_CLEAN_TIMEOUT then MyCount := 0;
    if _BShutdown then begin
      Break;
    end;
    Sleep(1000);
    if FAV_SYNCOBJECT_STATUS <> nil then begin
      FAV_SYNCOBJECT_STATUS.DumpToFile(MyCount=0);
    end;
  end;
  _BTerminated := True;
end;

constructor TThreadSyncObjectStatus.Create(CreateSuspended: Boolean;
  const StackSize: SizeUInt);
begin
  _BTerminated := True;
  _BShutdown := False;
  inherited Create(CreateSuspended, StackSize);
end;

destructor TThreadSyncObjectStatus.Destroy;
begin
  ShutDownThread;
  inherited Destroy;
end;

procedure TThreadSyncObjectStatus.ShutDownThread;
begin
  while _BTerminated = False do begin
    _BShutdown := True;
    Sleep(10)
  end;
end;

{ TFavMutexDebug }

constructor TFavMutexDebug.Create(InObjectName: String; InLocation: String;
  InName: String; InInitialValue: Boolean);
{$ifdef Windows}
var MySecurDesc: SECURITY_DESCRIPTOR;
    MyMutexAttributes: TSecurityAttributes;
{$endif}
{$ifdef Linux}
var MyMutexAttributes: pthread_mutexattr_t;
    MyRV: Longint;
{$endif}
begin
  FObjectName := InObjectName;
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_CREATE, GetThreadID, InObjectName, InLocation);
  end;


  _Name := InName;
  _InitialValue:= InInitialValue;
  {$ifdef Windows}
  if ( InitializeSecurityDescriptor (@MySecurDesc, SECURITY_DESCRIPTOR_REVISION )) then begin
    if ( SetSecurityDescriptorDacl (@MySecurDesc, TRUE, nil, FALSE) ) then begin
      // initialize security attributes
      MyMutexAttributes.nLength := sizeof ( SECURITY_ATTRIBUTES );
      MyMutexAttributes.lpSecurityDescriptor := @MySecurDesc;
      MyMutexAttributes.bInheritHandle := TRUE;
    end;
  end;
  if _Name = '' then begin
    FMutex := CreateMutex(@MyMutexAttributes, InInitialValue, nil);
  end
  else begin
    FMutex := CreateMutex(@MyMutexAttributes, InInitialValue, PChar(_Name));
  end;
  if FMutex = 0 then begin
    raise Exception.Create(GetError('CREATE', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if InName = '' then begin
    FillChar(MyMutexAttributes, SizeOf(MyMutexAttributes), 0);
    pthread_mutexattr_settype(@MyMutexAttributes, PTHREAD_MUTEX_RECURSIVE_NP);
    MyRv := pthread_mutex_init(@FMutex, @MyMutexAttributes);
    if (MyRv <> 0) then begin
      raise Exception.Create(GetError('INIT', MyRV));
    end;
  end
  else begin
    FMutexFile := FpOpen(PChar(_Name),
                  O_RDWR      +   // open the file for both read and write access
                  O_CREAT     +   // create file if it does not already exist
                  $02000000, //O_CLOEXEC   ,   // close on execute
                  S_IRUSR     +   // user permission: read
                  S_IWUSR     );  // user permission: write
    if FMutexFile <= 0 then begin
      raise Exception.Create(GetError('NAMEDMUTEX-CREATE', MyRV));
    end;
  end;
  {$endif}
end;

destructor TFavMutexDebug.Destroy;
{$ifdef Windows}
var MyRv: LongBool;
{$endif}
{$ifdef Linux}
var MyRV: Longint;
{$endif}
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_DESTROY, GetThreadID, FObjectName, 'unknown');
  end;
  {$ifdef Windows}
  MyRv := CloseHandle(FMutex);
  if MyRv = False then begin
    raise Exception.Create(GetError('DESTROY', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin;
    MyRv := pthread_mutex_destroy (@FMutex);
    if MyRv <> 0 then begin
      raise Exception.Create(GetError('DESTROY', MyRV));
    end;
  end
  else begin
    if FMutexFile <> 0 then begin;
      MyRv := FpFLock(FMutexFile,  LOCK_UN); // F_ULOCK, 0 );
      Fpclose(FMutexFile);
      if MyRv < 0 then begin
        raise Exception.Create(GetError('NAMEDMUTEX-DESTROY', fpgeterrno));
      end;
    end
    else begin
      raise Exception.Create(_LastErrorMsg);
    end;
  end;
  {$endif}
  inherited;
end;

procedure TFavMutexDebug.Release(InLocation: String);
{$ifdef Windows}
var MyRv: LongBool;
{$endif}
{$ifdef Linux}
var MyRv: Longint;
{$endif}
begin
  {$ifdef Windows}
  MyRv := ReleaseMutex(FMutex);
  if MyRV = False then begin
    raise Exception.Create(GetError('RELEASE', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRv := pthread_mutex_unlock (@FMutex);
//    _LastThreadID := 0;
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('LOCK', MyRv));
    end;
  end
  else begin
    if FMutexFile <> 0 then begin;
      MyRv := fpFlock(FMutexFile, LOCK_UN); //F_ULOCK, 0 );    F_ULOCK := 0
//      FpClose(FMutexFile);
      if MyRv < 0 then begin
        raise Exception.Create(GetError('NAMEDMUTEX-RELEASE', fpgeterrno));
      end;
    end
    else begin
      raise Exception.Create('Named Mutex: ' + _Name + ' not created');
    end;
  end
  {$endif}

  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_UNLOCK, GetThreadID, FObjectName, InLocation);
  end;
end;

procedure TFavMutexDebug.Wait(InLocation: String);
var MyRv: Longint;
{$ifdef Linux}
{$endif}
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_LOCK, GetThreadID, FObjectName, InLocation);
  end;

  {$ifdef Windows}
  MyRv := WaitForSingleObject(FMutex, Infinite);
  if MyRV <> WAIT_OBJECT_0 then begin
    raise Exception.Create(GetError('WAIT', MyRv));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRv := pthread_mutex_lock (@FMutex);
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('LOCK', MyRv));
    end;
  end
  else begin
    if FMutexFile <> 0 then begin;
      MyRv := fpFlock(FMutexFile, LOCK_SH); //LOCK_SH;// lock the "semaphore" F_LOCK - wait ifinitelly F_TLOCK - Try to lock
      if MyRv < 0 then begin
        raise Exception.Create(GetError('NAMEDMUTEX-WAIT', fpgeterrno));
      end;
    end
    else begin
      raise Exception.Create('Named Mutex: ' + _Name + ' not created');
    end;
  end;
  {$endif}
end;

function TFavMutexDebug.GetError(InAction: String; InError: Int64): String;
var MyResult: String;
begin
  MyResult := '';
  _LastErroNo:=InError;
{$ifdef Linux}
  if InAction = 'INIT' then begin
    if InError = EAGAIN then
      MyResult := 'The system lacked the necessary resources (other than memory) to initialise another mutex'
    else if InError = ENOMEM then
      MyResult := 'Insufficient memory exists to initialise the mutex.'
    else if InError = EPERM then
      MyResult := 'The caller does not have the privilege to perform the operation.'
    else if InError = EBUSY then
      MyResult := 'The implementation has detected an attempt to re-initialise the object referenced by mutex, a previously initialised, but not yet destroyed, mutex.'
    else if InError = EINVAL then
      MyResult := 'The value specified by attr is invalid.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'DESTROY' then begin
    if InError = EBUSY then
      MyResult := 'The implementation has detected an attempt to destroy the object referenced by mutex while it is locked or referenced (for example, while being used in a pthread_cond_wait() or pthread_cond_timedwait()) by another thread.'
    else if InError = EINVAL then
      MyResult := 'The value specified by mutex is invalid.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'UNLOCK' then begin
    if InError = EINVAL then
      MyResult := 'Mutex is not an initialized mutex. '
    else if InError = EFAULT then
      MyResult := 'Mutex is an invalid pointer. '
    else if InError = EPERM then
      MyResult := 'The calling thread does not own the mutex. '
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'LOCK' then begin
    if InError = EINVAL then
      MyResult := 'The mutex was created with the protocol attribute having the value PTHREAD_PRIO_PROTECT and the calling thread''s priority is higher than the mutex''s current priority ceiling. The value specified by mutex does not refer to an initialised mutex object.'
    else if InError = EBUSY then
      MyResult := 'The mutex could not be acquired because it was already locked.'
    else if InError = EFAULT then
      MyResult := 'mutex is an invalid pointer. '
    else if InError = EAGAIN then
      MyResult := 'The mutex could not be acquired because the maximum number of recursive locks for mutex has been exceeded.'
    else if InError = EDEADLK then
      MyResult := 'The current thread already owns the mutex.'
    else if InError = EPERM then
      MyResult := 'The calling thread does not own the mutex. '
    else
      MyResult := 'Unknown error';
  end
  else begin
    if InError = EACCES then
      MyResult := 'The file is locked and F_TLOCK or F_TEST was specified, or the operation is prohibited because the file has been memory-mapped by another process.'
    else if InError = EAGAIN then
      MyResult := 'The file is locked and F_TLOCK or F_TEST was specified, or the operation is prohibited because the file has been memory-mapped by another process.'
    else if InError = EBADF then
      MyResult := 'fd is not an open file descriptor; or cmd is F_LOCK or F_TLOCK and fd is not a writable file descriptor.'
    else if InError = EDEADLK then
      MyResult := 'The command was T_LOCK and this lock operation would cause a deadlock.'
    else if InError = EINVAL then
      MyResult := 'An invalid operation was specified in fd.'
    else if InError = ENOLCK then
      MyResult := 'Too many segment locks open, lock table is full.'
    else
      MyResult := 'Unknown error';
  end;
{$endif}
{$ifdef Windows}
  if (InAction = 'WAIT') or (InAction = 'RELEASE') then begin
    if InError = WAIT_ABANDONED then
      MyResult := 'The specified object is a mutex object that was not released by the thread that owned the mutex object before the owning thread terminated. Ownership of the mutex object is granted to the calling thread and the mutex state is set to nonsignaled. If the mutex was protecting persistent state information, you should check it for consistency.'
    else if InError = WAIT_OBJECT_0 then
      MyResult := 'The state of the specified object is signaled.'
    else if InError = WAIT_TIMEOUT then
      MyResult := 'The time-out interval elapsed, and the object''s state is nonsignaled.'
    else if InError = WAIT_FAILED then
      MyResult := 'The function has failed. To get extended error information, call GetLastError: ' + IntToStr(GetLastError)
    else
      MyResult := 'Unknown error';
  end
  else if (InAction = 'CREATE') or (InAction = 'DESTROY') then begin
    MyResult := 'The function has failed. To get extended error information, call GetLastError: ' + IntToStr(GetLastError);
  end;
{$endif}

  MyResult := 'Mutex: ' + _Name + '; Action: ' + InAction + '; Rv: ' + IntToStr(InError) + '; ' + MyResult;
  _LastErrorMsg:=MyResult;

  Result := MyResult;
end;

{ TFavSyncObjectsStatuses }

constructor TFavSyncObjectsStatuses.Create;
begin
  _List := TList.Create;
  _Controller := TFavMutex.Create();

  try
    if FileExists('SyncObjects.err') then begin
      DeleteFile('SyncObjects.err');
    end;
  except
  end;
end;

destructor TFavSyncObjectsStatuses.Destroy;
begin
  DumpToFile(false);
  DeleteAllItems;
  FreeAndNil(_List);
  FreeAndNil(_Controller);
  inherited Destroy;
end;

procedure TFavSyncObjectsStatuses.RegisterAction(InAction: Integer;
  InThreadID: Int64; InName: String; InLocation: String);
var MyCls: TFavSyncObjectsStatus;
    MyIHaveController: Boolean;
    MyError: String;
begin
  if InName <> '' then begin
    MyIHaveController := False;
    MyError := '';
    try
      _Controller.Wait;
      MyIHaveController := True;
      MyCls := FindItem(InName, InThreadID);
      if MyCls = nil then begin
        MyCls := TFavSyncObjectsStatus.Create;
        _List.Add(MyCls);
      end;
      MyCls._data.Name := InName;
      MyCls._data.ThreadID := InThreadID;

      with MyCls do begin
        if InAction = FAV_SEM_CREATE then begin
          _data.CreateCnt := _data.CreateCnt + 1;
          _data.CreateLocation := InLocation;
          _data.CreateTime := Now;
        end
        else if InAction = FAV_SEM_READ_LOCK then begin
          _data.ReadCnt := _data.ReadCnt + 1;
          _data.ReadLocation := InLocation;
          _data.ReadTime := Now;
        end
        else if InAction = FAV_SEM_READ_UNLOCK then begin
          _data.ReadUnlockCnt := _data.ReadUnlockCnt + 1;
          _data.ReadUnlockLocation := InLocation;
          _data.ReadUnlockTime := Now;
        end
        else if InAction = FAV_SEM_WRITE_LOCK then begin
          _data.WriteCnt := _data.WriteCnt + 1;
          _data.WriteLocation := InLocation;
          _data.WriteTime := Now;
        end
        else if InAction = FAV_SEM_WRITE_UNLOCK then begin
          _data.WriteUnlockCnt := _data.WriteUnlockCnt + 1;
          _data.WriteUnlockLocation := InLocation;
          _data.WriteUnlockTime := Now;
        end
        else if InAction = FAV_SEM_DESTROY then begin
          _data.DestroyCnt := _data.DestroyCnt + 1;
          _data.DestroyLocation := InLocation;
          _data.DestroyTime := Now;
        end;


        if InAction = FAV_SEM_READ_LOCK then begin
          if _data.ReadCnt - _data.ReadUnlockCnt > 1 then MyError := MyError + ('ReadLock: ' + InName + ', ' + IntToStr(InThreadID) + ', ' + InLocation + ' = ReadCnt - ReadUnlockCnt > 1') + #13#10;
          if _data.WriteCnt - _data.WriteUnlockCnt > 1 then MyError := MyError + ('ReadLock: ' + InName + ', ' + IntToStr(InThreadID) + ', ' + InLocation + ' = WriteCnt - WriteUnlockCnt > 1. You Alread have right lock') + #13#10;
        end
        else if InAction = FAV_SEM_READ_UNLOCK then begin
          if _data.ReadCnt < _data.ReadUnlockCnt then MyError := MyError + ('ReadUnlock: ' + InName + ', ' + IntToStr(InThreadID) + ', ' + InLocation + ' = ReadCnt < ReadUnlockCnt') + #13#10;
        end
        else if InAction = FAV_SEM_WRITE_LOCK then begin
          if _data.WriteCnt - _data.WriteUnlockCnt > 1 then MyError := MyError + ('WriteLock: ' + InName + ', ' + IntToStr(InThreadID) + ', ' + InLocation + ' = WriteCnt - WriteUnlockCnt > 1') + #13#10;
          if _data.ReadCnt - _data.ReadUnlockCnt > 1 then MyError := MyError + ('WriteLock: ' + InName + ', ' + IntToStr(InThreadID) + ', ' + InLocation + ' = ReadCnt - ReadUnlockCnt > 1. You already have ReadLock') + #13#10;
        end
        else if InAction = FAV_SEM_WRITE_UNLOCK then begin
          if _data.WriteCnt < _data.WriteUnlockCnt then MyError := MyError + ('WriteUnlock: ' + InName + ', ' + IntToStr(InThreadID) + ', ' + InLocation + ' = WriteCnt < WriteUnlockCnt') + #13#10;
        end;

        if MyError <> '' then begin
          AppendToFile('SyncObjects.err', FormatDateTime('yyyy-mm-dd hh:nn:ss.zzz', Now) + ': ' + MyError);
          AppendToFile('SyncObjects.err', GenerateDumpList);
          if FAV_RAISE_SYNCOBJECTS_REGISTRATION_ERROR then raise Exception.create(MyError);
        end;
      end;

      if MyIHaveController then begin
        _Controller.Release;
        MyIHaveController := False;
      end;
    except
      if MyIHaveController then begin
        _Controller.Release;
        MyIHaveController := False;
      end;
      raise;
    end;
  end;
end;

procedure TFavSyncObjectsStatuses.DeleteAllItems;
var F: Integer;
    MyCls: TFavSyncObjectsStatus;
begin
  try
    _Controller.Wait;
    for F := 0 to _list.Count-1 do begin
      MyCls := TFavSyncObjectsStatus(_list.Items[F]);
      FreeAndNil(MyCls);
    end;
    _list.Clear;
  finally
    _Controller.Release;
  end;
end;

function TFavSyncObjectsStatuses.FindItem(InName: String; InThreadID: Int64
  ): TFavSyncObjectsStatus;
var F: Integer;
    MyCls, MyResult: TFavSyncObjectsStatus;
begin
  MyCls := nil;
  MyResult := nil;
  for F := 0 to _List.Count-1 do begin
    MyCls := TFavSyncObjectsStatus(_list.Items[F]);
    if (MyCls._data.Name = InName) and (MyCls._data.ThreadID = InThreadID) then begin
      MyResult := MyCls;
      Break;
    end;
  end;
  Result := MyResult;
end;

function TFavSyncObjectsStatuses.DateTimeToString(InDateTime: TDateTime
  ): String;
var MyResult: String;
begin
  MyResult := '';
  if InDateTime <> 0 then begin
    MyResult := FormatDateTime('yyyymmdd_hhnnss.zzz', InDateTime);
  end;
  Result := MyResult;
end;

function TFavSyncObjectsStatuses.GenerateDumpList: String;
var MyCls: TFavSyncObjectsStatus;
    MyString: String;
    F: Integer;
    MyWaitReaders: Integer;
    MyWaitWriters: Integer;
    MyCntCreate, MyCntDestroy, MyCntRL, MyCntRU, MyCntWL, MyCntWU: Int64;
begin
  MyString := '';
  MyWaitReaders := 0;
  MyWaitWriters := 0;
  MyCntCreate  := 0;
  MyCntDestroy := 0;
  MyCntRL := 0;
  MyCntRU := 0;
  MyCntWL := 0;
  MyCntWU := 0;
  MyString := MyString + Format('%-60.60s, %8.8s, |CREATE: %6.0s, %-20.20s, %-50.50s |RL: %6.0s, %-20.20s, %-50.50s |RU: %6.0s, %-20.20s, %-50.50s |WL: %6.0s, %-20.20s, %-50.50s |WU: %6.0s, %-20.20s, %-50.50s |DESTROY: %6.0s, %-20.20s, %-50.50s',
                                [
                                'ObjectName',
                                'ThreadID',
                                'CreateCnt',
                                'CreateTime',
                                'CreateLocation',

                                'ReadCnt',
                                'ReadTime',
                                'ReadLocation',
                                'ReadUnlockCnt',
                                'ReadUnlockTime',
                                'ReadUnlockLocation',

                                'WriteCnt',
                                'WriteTime',
                                'WriteLocation',
                                'WriteUnlockCnt',
                                'WriteUnlockTime',
                                'WriteUnlockLocation',

                                'DestroyCnt',
                                'DestroyTime',
                                'DestroyLocation']) + #13#10;

  for F := 0 to _List.Count-1 do begin
    MyCls := TFavSyncObjectsStatus(_list.Items[F]);
    with MyCls._data do begin
      if MyCls._data.ReadCnt > MyCls._data.ReadUnlockCnt then MyWaitReaders := MyWaitReaders + 1;
      if MyCls._data.WriteCnt > MyCls._data.WriteUnlockCnt then MyWaitWriters := MyWaitWriters + 1;
      MyCntCreate := MyCntCreate + MyCls._data.CreateCnt;
      MyCntDestroy := MyCntDestroy + MyCls._data.DestroyCnt;
      MyCntRL := MyCntRL + MyCls._data.ReadCnt;
      MyCntRU := MyCntRU + MyCls._data.ReadUnlockCnt;
      MyCntWL := MyCntWL + MyCls._data.WriteCnt;
      MyCntWU := MyCntWU + MyCls._data.WriteUnlockCnt;

      MyString := MyString + Format('%-60.60s, %8.8d, |CREATE: %6.0d, %-20.20s, %-50.50s |RL: %6.0d, %-20.20s, %-50.50s |RU: %6.0d, %-20.20s, %-50.50s |WL: %6.0d, %-20.20s, %-50.50s |WU: %6.0d, %-20.20s, %-50.50s |DESTROY: %6.0d, %-20.20s, %-50.50s',
                          [
                          Name,
                          (ThreadID),
                          (CreateCnt),
                          DateTimeToString(CreateTime),
                          CreateLocation,

                          (ReadCnt),
                          DateTimeToString(ReadTime),
                          ReadLocation,
                          (ReadUnlockCnt),
                          DateTimeToString(ReadUnlockTime),
                          ReadUnlockLocation,

                          (WriteCnt),
                          DateTimeToString(WriteTime),
                          WriteLocation,
                          (WriteUnlockCnt),
                          DateTimeToString(WriteUnlockTime),
                          WriteUnlockLocation,

                          (DestroyCnt),
                          DateTimeToString(DestroyTime),
                          DestroyLocation]) + #13#10;
    end;
  end;
  MyString := MyString + Format('Line Count: %d', [_List.Count]) + #13#10;
  MyString := MyString + Format('Waiting Readers: %d, Waiting Writers: %d', [MyWaitReaders, MyWaitWriters]) + #13#10;
  MyString := MyString + Format('CntCreate: %d, CntDestroy: %d, ReadLock: %d, ReadUnLock: %d, WriteLock: %d, WriteUnLock: %d', [MyCntCreate, MyCntDestroy, MyCntRL, MyCntRU, MyCntWL, MyCntWU]) + #13#10;
  Result := MyString;
end;

procedure TFavSyncObjectsStatuses.SaveToFile(InFileName: String; InString: String);
var MyFile: Text;
    MyFileAssigned: Boolean;
    MyFileOpened: Boolean;
begin
  MyFileAssigned:=False;
  MyFileOpened:=False;
  try
    AssignFile(MyFile, InFileName);
    MyFileAssigned:=True;
    Rewrite(MyFile);
    MyFileOpened := True;

    if InString <> '' then begin
      Write(MyFile, InString);
      Flush(MyFile);
    end;

    CloseFile(MyFile);
    MyFileOpened:=False;
    MyFileAssigned:=False;
  except
    on E: Exception do begin
      if MyFileAssigned then begin
        try
          if MyFileOpened then begin
            CloseFile(MyFile);
          end;
        except
        end;
      end;
    end;
  end;
end;

procedure TFavSyncObjectsStatuses.AppendToFile(InFileName: String;
  InString: String);
var MyFile: Text;
    MyFileAssigned: Boolean;
    MyFileOpened: Boolean;
begin
  MyFileAssigned:=False;
  MyFileOpened := False;
  try
    AssignFile(MyFile, InFileName);
    MyFileAssigned:=True;
    if FileExists(InFileName) then begin
      Append(MyFile);
      MyFileOpened := True;
    end
    else begin
      Rewrite(MyFile);
      MyFileOpened := True;
    end;
    Write(MyFile, InString);
    Flush(MyFile);
    CloseFile(MyFile);
    MyFileOpened := False;
    MyFileAssigned:=False;
  except
    on E: Exception do begin
      if MyFileAssigned then begin
        try
          if MyFileOpened then begin
            CloseFile(MyFile);
          end;
        except
        end;
      end;
      raise;
    end;
  end;
end;

procedure TFavSyncObjectsStatuses.CleanSynchronizedLines;
var MyCls: TFavSyncObjectsStatus;
    F: Integer;
begin
  F := 0;
  while F < _list.Count do begin
    MyCls := TFavSyncObjectsStatus(_list.Items[F]);
    with MyCls._data do begin
      if (MyCls._data.ReadCnt = MyCls._data.ReadUnlockCnt) and
         (MyCls._data.WriteCnt = MyCls._data.WriteUnlockCnt) and
         (MyCls._data.CreateCnt = MyCls._data.DestroyCnt) // remove if it is not created or it is created and destroyed
      then begin
        // this line is synchronized and it will be removed
        _list.Remove(MyCls);
        FreeAndNil(MyCls);
      end
      else begin
        // nothing to remove
        F := F + 1;
      end;
    end;
  end;
end;

procedure TFavSyncObjectsStatuses.DumpToFile(InCleanList: Boolean);
var MyIHaveMutex: Boolean;
    MyString: String;
begin
  try
    _Controller.Wait;
    MyIHaveMutex := True;

    if InCleanList then begin
      //CleanSynchronizedLines;
    end;

    MyString := GenerateDumpList;

    if MyString <> '' then begin
      SaveToFile('SyncObjects.dmp', MyString);
    end;

    if MyIHaveMutex then begin
      _Controller.Release;
      MyIHaveMutex := False;
    end;
  except
    if MyIHaveMutex then begin
      _Controller.Release;
      MyIHaveMutex := False;
    end;
  end;

end;

{ TFavSyncObjectsStatus }

constructor TFavSyncObjectsStatus.Create;
begin
  _data := InitTFavRecSyncObjectStatus;
end;

destructor TFavSyncObjectsStatus.Destroy;
begin
  _data := InitTFavRecSyncObjectStatus;
  inherited Destroy;
end;

{ TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi }

constructor TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi.Create(
  InObjectName: String; InLocation: String; InMaxReaders: Integer);
begin
  inherited Create;
  FObjectName := InObjectName;
  FMaxReaders := InMaxReaders;
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_CREATE, GetThreadID, InObjectName, InLocation);
  end;


  {create the primitive synchronization objects}
  FBlockedReaders := TFavSemaphore.Create('', 0, FMaxReaders);
  FBlockedWriters := TFavSemaphore.Create('', 0, 1);
  FController := TFavMutex.Create();
end;

destructor TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi.Destroy;
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_DESTROY, GetThreadID, FObjectName, 'unknown');
  end;

  FreeAndNil(FBlockedReaders);
  FreeAndNil(FBlockedWriters);
  FreeAndNil(FController);

  inherited Destroy;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi.WaitForRead(
  InLocation: String);
var HaveToWait : boolean;
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_READ_LOCK, GetThreadID,FObjectName, InLocation);
  end;

  {acquire the controlling critical section}
  FController.Wait;

  {if there is a writer executing or there is at least one writer
   waiting, add ourselves as a waiting reader, make sure we wait}
  if FActiveWriter or (FWaitingWriters <> 0) then begin
    Assert(FWaitingReaders < FMaxReaders, 'TtdReadWriteSync: maximum waiting readers reached');
    inc(FWaitingReaders);
    HaveToWait := true;
  end

  {otherwise, add ourselves as another executing reader,
   and make sure we don't wait}
  else begin
    inc(FActiveReaders);
    HaveToWait := false;
  end;

  {release the controlling critical section}
  FController.Release;

  {if we have to wait, then do so}
  if HaveToWait then begin
    FBlockedReaders.Wait;
  end;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi.WaitForWrite(
  InLocation: String);
var HaveToWait : boolean;
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_LOCK, GetThreadID,FObjectName, InLocation);
  end;

  {acquire the controlling critical section}
  FController.Wait;

  {if there is another writer running or there are active readers, add
   ourselves as a waiting writer, and make sure we wait}
  if FActiveWriter or (FActiveReaders <> 0) then begin
    inc(FWaitingWriters);
    HaveToWait := true;
  end

  {otherwise, add ourselves as another executing writer, and make sure
   we don't wait}
  else begin
    FActiveWriter := true;
    HaveToWait := false;
  end;

  {release the controlling critical section}
  FController.Release;

  {if we have to wait, then do so}
  if HaveToWait then
    FBlockedWriters.Wait;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi.ReleaseRead(
  InLocation: String);
begin
  {acquire the controlling critical section}
  FController.Wait;

  {we've finished reading}
  dec(FActiveReaders);

  {if we are the last reader reading and there is at
   least one writer waiting, then release it}
  if (FActiveReaders = 0) and (FWaitingWriters <> 0) then begin
    dec(FWaitingWriters);
    FActiveWriter := true;
    FBlockedWriters.Release;
  end;

  {release the controlling critical section}
  FController.Release;

  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_READ_UNLOCK, GetThreadID, FObjectName, InLocation);
  end;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityTomesOfDelphi.ReleaseWrite(
  InLocation: String);
begin
  {acquire the controlling critical section}
  FController.Wait;

  {we've finished writing}
  FActiveWriter := false;

  {if there is at least one reader waiting then release them all}
  if (FWaitingReaders <> 0) then begin
    FActiveReaders := FWaitingReaders;
    FWaitingReaders := 0;
    FBlockedReaders.Release(FActiveReaders);
  end

  {otherwise, if there is at least one waiting writer, release one}
  else if (FWaitingWriters <> 0) then begin
    dec(FWaitingWriters);
    FActiveWriter := true;
    FBlockedWriters.Release;
  end;

  {release the controlling critical section}
  FController.Release;

  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_UNLOCK, GetThreadID, FObjectName, InLocation);
  end;
end;

{ TFavWriterReaderSemaphoreWriterPriority }

constructor TFavWriterReaderSemaphoreWriterPriorityFPC.Create(
  InObjectName: String; InLocation: String);
begin
  inherited Create;
  _sync := TMultiReadExclusiveWriteSynchronizer.Create;

  FObjectName := InObjectName;
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_CREATE, GetThreadID, InObjectName, InLocation);
  end;
end;

destructor TFavWriterReaderSemaphoreWriterPriorityFPC.Destroy;
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_DESTROY, GetThreadID, FObjectName, 'unknown');
  end;

  FreeAndNil(_sync);
  inherited Destroy;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityFPC.WaitForRead(
  InLocation: String);
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_READ_LOCK, GetThreadID, FObjectName, InLocation);
  end;
  _sync.Beginread;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityFPC.WaitForWrite(
  InLocation: String);
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_LOCK, GetThreadID, FObjectName, InLocation);
  end;
  _sync.Beginwrite;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityFPC.ReleaseRead(
  InLocation: String);
begin
  _sync.Endread;

  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_READ_UNLOCK, GetThreadID, FObjectName, InLocation);
  end;
end;

procedure TFavWriterReaderSemaphoreWriterPriorityFPC.ReleaseWrite(
  InLocation: String);
begin
  _sync.Endwrite;

  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_UNLOCK, GetThreadID, FObjectName, InLocation);
  end;
end;

{ TFavSwitch }

constructor TFavSwitch.Create;
begin
  inherited Create;
  _Value := False;
end;

destructor TFavSwitch.Destroy;
begin
  inherited Destroy;
end;

{ TFavMutex }

constructor TFavMutex.Create(InName: String; InInitialValue: Boolean);
{$ifdef Windows}
var MySecurDesc: SECURITY_DESCRIPTOR;
    MyMutexAttributes: TSecurityAttributes;
{$endif}
{$ifdef Linux}
var MyMutexAttributes: pthread_mutexattr_t;
    MyRV: Longint;
{$endif}
begin
  _Name := InName;
  _InitialValue:= InInitialValue;
  {$ifdef Windows}
  if ( InitializeSecurityDescriptor (@MySecurDesc, SECURITY_DESCRIPTOR_REVISION )) then begin
    if ( SetSecurityDescriptorDacl (@MySecurDesc, TRUE, nil, FALSE) ) then begin
      // initialize security attributes
      MyMutexAttributes.nLength := sizeof ( SECURITY_ATTRIBUTES );
      MyMutexAttributes.lpSecurityDescriptor := @MySecurDesc;
      MyMutexAttributes.bInheritHandle := TRUE;
    end;
  end;
  if _Name = '' then begin
    FMutex := CreateMutex(@MyMutexAttributes, InInitialValue, nil);
  end
  else begin
    FMutex := CreateMutex(@MyMutexAttributes, InInitialValue, PChar(_Name));
  end;
  if FMutex = 0 then begin
    raise Exception.Create(GetError('CREATE', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if InName = '' then begin
    FillChar(MyMutexAttributes, SizeOf(MyMutexAttributes), 0);
    pthread_mutexattr_settype(@MyMutexAttributes, PTHREAD_MUTEX_RECURSIVE_NP);
    MyRv := pthread_mutex_init(@FMutex, @MyMutexAttributes);
    if (MyRv <> 0) then begin
      raise Exception.Create(GetError('INIT', MyRV));
    end;
  end
  else begin
    FMutexFile := FpOpen(PChar(_Name),
                  O_RDWR      +   // open the file for both read and write access
                  O_CREAT     +   // create file if it does not already exist
                  $02000000, //O_CLOEXEC   ,   // close on execute
                  S_IRUSR     +   // user permission: read
                  S_IWUSR     );  // user permission: write
    if FMutexFile <= 0 then begin
      raise Exception.Create(GetError('NAMEDMUTEX-CREATE', MyRV));
    end;
  end;
  {$endif}
end;

destructor TFavMutex.Destroy;
{$ifdef Windows}
var MyRv: LongBool;
{$endif}
{$ifdef Linux}
var MyRV: Longint;
{$endif}
begin
  {$ifdef Windows}
  MyRv := CloseHandle(FMutex);
  if MyRv = False then begin
    raise Exception.Create(GetError('DESTROY', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin;
    MyRv := pthread_mutex_destroy (@FMutex);
    if MyRv <> 0 then begin
      raise Exception.Create(GetError('DESTROY', MyRV));
    end;
  end
  else begin
    if FMutexFile <> 0 then begin;
      MyRv := FpFLock(FMutexFile,  LOCK_UN); // F_ULOCK, 0 );
      Fpclose(FMutexFile);
      if MyRv < 0 then begin
        raise Exception.Create(GetError('NAMEDMUTEX-DESTROY', fpgeterrno));
      end;
    end
    else begin
      raise Exception.Create(_LastErrorMsg);
    end;
  end;
  {$endif}
  inherited;
end;

procedure TFavMutex.Release;
{$ifdef Windows}
var MyRv: LongBool;
{$endif}
{$ifdef Linux}
var MyRv: Longint;
{$endif}
begin
  {$ifdef Windows}
  MyRv := ReleaseMutex(FMutex);
  if MyRV = False then begin
    raise Exception.Create(GetError('RELEASE', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRv := pthread_mutex_unlock (@FMutex);
//    _LastThreadID := 0;
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('LOCK', MyRv));
    end;
  end
  else begin
    if FMutexFile <> 0 then begin;
      MyRv := fpFlock(FMutexFile, LOCK_UN); //F_ULOCK, 0 );    F_ULOCK := 0
//      FpClose(FMutexFile);
      if MyRv < 0 then begin
        raise Exception.Create(GetError('NAMEDMUTEX-RELEASE', fpgeterrno));
      end;
    end
    else begin
      raise Exception.Create('Named Mutex: ' + _Name + ' not created');
    end;
  end
  {$endif}
end;

procedure TFavMutex.Wait;
var MyRv: Longint;
{$ifdef Linux}
{$endif}
begin
  {$ifdef Windows}
  MyRv := WaitForSingleObject(FMutex, Infinite);
  if MyRV <> WAIT_OBJECT_0 then begin
    raise Exception.Create(GetError('WAIT', MyRv));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRv := pthread_mutex_lock (@FMutex);
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('LOCK', MyRv));
    end;
  end
  else begin
    if FMutexFile <> 0 then begin;
      MyRv := fpFlock(FMutexFile, LOCK_SH); //LOCK_SH;// lock the "semaphore" F_LOCK - wait ifinitelly F_TLOCK - Try to lock
      if MyRv < 0 then begin
        raise Exception.Create(GetError('NAMEDMUTEX-WAIT', fpgeterrno));
      end;
    end
    else begin
      raise Exception.Create('Named Mutex: ' + _Name + ' not created');
    end;
  end;
  {$endif}
end;

function TFavMutex.GetError(InAction: String; InError: Int64): String;
var MyResult: String;
begin
  MyResult := '';
  _LastErroNo:=InError;
{$ifdef Linux}
  if InAction = 'INIT' then begin
    if InError = EAGAIN then
      MyResult := 'The system lacked the necessary resources (other than memory) to initialise another mutex'
    else if InError = ENOMEM then
      MyResult := 'Insufficient memory exists to initialise the mutex.'
    else if InError = EPERM then
      MyResult := 'The caller does not have the privilege to perform the operation.'
    else if InError = EBUSY then
      MyResult := 'The implementation has detected an attempt to re-initialise the object referenced by mutex, a previously initialised, but not yet destroyed, mutex.'
    else if InError = EINVAL then
      MyResult := 'The value specified by attr is invalid.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'DESTROY' then begin
    if InError = EBUSY then
      MyResult := 'The implementation has detected an attempt to destroy the object referenced by mutex while it is locked or referenced (for example, while being used in a pthread_cond_wait() or pthread_cond_timedwait()) by another thread.'
    else if InError = EINVAL then
      MyResult := 'The value specified by mutex is invalid.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'UNLOCK' then begin
    if InError = EINVAL then
      MyResult := 'Mutex is not an initialized mutex. '
    else if InError = EFAULT then
      MyResult := 'Mutex is an invalid pointer. '
    else if InError = EPERM then
      MyResult := 'The calling thread does not own the mutex. '
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'LOCK' then begin
    if InError = EINVAL then
      MyResult := 'The mutex was created with the protocol attribute having the value PTHREAD_PRIO_PROTECT and the calling thread''s priority is higher than the mutex''s current priority ceiling. The value specified by mutex does not refer to an initialised mutex object.'
    else if InError = EBUSY then
      MyResult := 'The mutex could not be acquired because it was already locked.'
    else if InError = EFAULT then
      MyResult := 'mutex is an invalid pointer. '
    else if InError = EAGAIN then
      MyResult := 'The mutex could not be acquired because the maximum number of recursive locks for mutex has been exceeded.'
    else if InError = EDEADLK then
      MyResult := 'The current thread already owns the mutex.'
    else if InError = EPERM then
      MyResult := 'The calling thread does not own the mutex. '
    else
      MyResult := 'Unknown error';
  end
  else begin
    if InError = EACCES then
      MyResult := 'The file is locked and F_TLOCK or F_TEST was specified, or the operation is prohibited because the file has been memory-mapped by another process.'
    else if InError = EAGAIN then
      MyResult := 'The file is locked and F_TLOCK or F_TEST was specified, or the operation is prohibited because the file has been memory-mapped by another process.'
    else if InError = EBADF then
      MyResult := 'fd is not an open file descriptor; or cmd is F_LOCK or F_TLOCK and fd is not a writable file descriptor.'
    else if InError = EDEADLK then
      MyResult := 'The command was T_LOCK and this lock operation would cause a deadlock.'
    else if InError = EINVAL then
      MyResult := 'An invalid operation was specified in fd.'
    else if InError = ENOLCK then
      MyResult := 'Too many segment locks open, lock table is full.'
    else
      MyResult := 'Unknown error';
  end;
{$endif}
{$ifdef Windows}
  if (InAction = 'WAIT') or (InAction = 'RELEASE') then begin
    if InError = WAIT_ABANDONED then
      MyResult := 'The specified object is a mutex object that was not released by the thread that owned the mutex object before the owning thread terminated. Ownership of the mutex object is granted to the calling thread and the mutex state is set to nonsignaled. If the mutex was protecting persistent state information, you should check it for consistency.'
    else if InError = WAIT_OBJECT_0 then
      MyResult := 'The state of the specified object is signaled.'
    else if InError = WAIT_TIMEOUT then
      MyResult := 'The time-out interval elapsed, and the object''s state is nonsignaled.'
    else if InError = WAIT_FAILED then
      MyResult := 'The function has failed. To get extended error information, call GetLastError: ' + IntToStr(GetLastError)
    else
      MyResult := 'Unknown error';
  end
  else if (InAction = 'CREATE') or (InAction = 'DESTROY') then begin
    MyResult := 'The function has failed. To get extended error information, call GetLastError: ' + IntToStr(GetLastError);
  end;
{$endif}

  MyResult := 'Mutex: ' + _Name + '; Action: ' + InAction + '; Rv: ' + IntToStr(InError) + '; ' + MyResult;
  _LastErrorMsg:=MyResult;

  Result := MyResult;
end;

{ TFavQueue }

constructor TFavQueue.Create(InMaxMsgNo: Longint);
begin
  inherited Create;
  _maxMsgNo:=InMaxMsgNo;
  _list := TList.Create;

  _mutexAccess := TFavMutex.Create();
  _semaphoreReaders := TFavSemaphore.Create('', 0, _maxMsgNo);
  _semaphoreWriters := TFavSemaphore.Create('', _maxMsgNo, _maxMsgNo);
end;

destructor TFavQueue.Destroy;
begin
  CleanQueue;
  if Assigned(_List) then FreeAndNil(_List);
  if Assigned(_mutexAccess) then FreeAndNil(_mutexAccess);
  if Assigned(_semaphoreReaders) then FreeAndNil(_semaphoreReaders);
  if Assigned(_semaphoreWriters) then FreeAndNil(_semaphoreWriters);
  inherited Destroy;
end;

procedure TFavQueue.PutData(var InString: String);
var MyData: TFavQueueData;
begin
  try
    _semaphoreWriters.wait;
    _mutexAccess.Wait;
    MyData := TFavQueueData.Create(InString);
    _list.Add(MyData);
    _mutexAccess.Release;
    _semaphoreReaders.Release;
  except
    on E: Exception do begin
      _semaphoreReaders.Release;
      _semaphoreWriters.wait;
      _mutexAccess.Release;
    end;
  end;
end;

function TFavQueue.GetData: String;
var MyData: TFavQueueData;
    MyResult: String;
begin
  MyResult := '';
  try
    _semaphoreReaders.wait;
    _mutexAccess.Wait;
    if _List.Count > 0 then begin
      MyData := TFavQueueData(_list.Items[0]);
      MyResult := MyData._data;
      FreeAndNil(MyData);
      _list.Delete(0);
    end;
    _mutexAccess.Release;
    _semaphoreWriters.Release;
  except
    on E: Exception do begin
      _semaphoreReaders.Release;
      _semaphoreWriters.Wait;
      _mutexAccess.Release;
      raise;
    end;
  end;
  Result := MyResult;
end;

procedure TFavQueue.CleanQueue;
var F: Longint;
    MyData: TFavQueueData;
begin
  try
    _semaphoreWriters.wait;
    _mutexAccess.Wait;
    for F := 0 to _list.Count-1 do begin
      MyData := TFavQueueData(_list.Items[F]);
      if Assigned(MyData) then FreeAndNil(MyData);
    end;
    _list.Clear;
    _mutexAccess.Release;
    _semaphoreReaders.Release;
  except
    on E: Exception do begin
      _semaphoreReaders.Release;
      _semaphoreWriters.wait;
      _mutexAccess.Release;
    end;
  end;
end;

{ TFavQueueData }

constructor TFavQueueData.Create(InData: String);
begin
  inherited Create;
  _data := InData;
end;

destructor TFavQueueData.Destroy;
begin
  _data:='';
  inherited Destroy;
end;

{ TFavWriterReaderSemaphoreSimple }

constructor TFavWriterReaderSemaphoreSimple.Create(InObjectName: String;
  InLocation: String);
begin
  inherited Create;
  FObjectName := InObjectName;
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_CREATE, GetThreadID, InObjectName, InLocation);
  end;

  _mutexMain := TFavMutex.Create();
end;

destructor TFavWriterReaderSemaphoreSimple.Destroy;
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_DESTROY, GetThreadID, FObjectName, 'unknown');
  end;
  _mutexMain.Free;
  inherited Destroy;
end;

procedure TFavWriterReaderSemaphoreSimple.WaitForRead(InLocation: String);
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_READ_LOCK, GetThreadID, FObjectName, InLocation);
  end;
  _mutexMain.Wait;
end;

procedure TFavWriterReaderSemaphoreSimple.WaitForWrite(InLocation: String);
begin
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_LOCK, GetThreadID, FObjectName, InLocation);
  end;
  _mutexMain.Wait;
end;

procedure TFavWriterReaderSemaphoreSimple.ReleaseRead(InLocation: String);
begin
  _mutexMain.Release;
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_READ_UNLOCK, GetThreadID, FObjectName, InLocation);
  end;
end;

procedure TFavWriterReaderSemaphoreSimple.ReleaseWrite(InLocation: String);
begin
  _mutexMain.Release;
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FAV_SYNCOBJECT_STATUS.RegisterAction(FAV_SEM_WRITE_UNLOCK, GetThreadID, FObjectName, InLocation);
  end;
end;

{ TFavSemaphore }
constructor TFavSemaphore.Create(InName: String; InInitialValue: Cardinal;
  InMaxValue: Cardinal);
{$ifdef Linux}
var MyRV: Longint;
{$endif}
{$ifdef Windows}
var MySecurDesc: SECURITY_DESCRIPTOR;
    MySemphoreAttributes: TSecurityAttributes;
{$endif}
begin
  _LastErroNo:=0;
  _LastErrorMsg:='';
  _Name:=InName;
  _InitialValue:= InInitialValue;
  _MaxValue:=InMaxValue;
  {$ifdef Windows}
  if ( InitializeSecurityDescriptor (@MySecurDesc, SECURITY_DESCRIPTOR_REVISION )) then begin
    if ( SetSecurityDescriptorDacl (@MySecurDesc, TRUE, nil, FALSE) ) then begin
      // initialize security attributes
      MySemphoreAttributes.nLength := sizeof ( SECURITY_ATTRIBUTES );
      MySemphoreAttributes.lpSecurityDescriptor := @MySecurDesc;
      MySemphoreAttributes.bInheritHandle := TRUE;
    end;
  end;
  if _Name = '' then begin
    FSemaphore := CreateSemaphore(@MySemphoreAttributes, InInitialValue, InMaxValue, nil);
  end
  else begin
    FSemaphore := CreateSemaphore(@MySemphoreAttributes, InInitialValue, InMaxValue, PChar(_Name));
  end;
  if FSemaphore = 0 then begin
    raise Exception.Create(GetError('CREATE', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRV := sem_init(@FSemaphore, 0, InInitialValue);
    if MyRV <> 0 then begin
      raise Exception.create(GetError('SEM_INIT', fpgeterrno));
    end;
  end
  else begin
    PSemaphore := sem_open(PChar(_Name), O_CREAT, S_IRUSR + S_IWUSR + S_IRGRP + S_IWGRP + S_IROTH + S_IWOTH, InInitialValue);
    if PSemaphore = SEM_FAILED then begin
      raise Exception.create(GetError('SEM_OPEN', fpgeterrno));
    end;
  end;
  {$endif}
end; // constructor TFavSemaphore.Create

destructor TFavSemaphore.Destroy;
{$ifdef Windows}
var MyRV: LongBool;
{$endif}
{$ifdef Linux}
var MyRV: Longint;
{$endif}
begin
  {$ifdef Windows}
  if FSemaphore <> 0 then begin
    MyRv := CloseHandle(FSemaphore);
    if MyRv = False then begin
      raise Exception.Create(GetError('DESTROY', GetLastError));
    end;
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRV := sem_destroy(@FSemaphore);
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('SEM_DESTROY', fpgeterrno));
    end;
  end
  else begin
    if PSemaphore <> nil then begin
      // why unlink - must test under linux
      MyRV := sem_unlink(PChar(_Name));
      if MyRV <> 0 then begin
        //raise Exception.create(GetError('SEM_UNLINK', LibC.fpgeterrno));
      end;

      MyRV := sem_close(PSemaphore);
      if MyRV <> 0 then begin
        raise Exception.Create(GetError('SEM_CLOSE', fpgeterrno));
      end;
    end
    else begin
      raise Exception.Create(_LastErrorMsg);
    end;
  end;
  {$endif}
  inherited;
end; // destructor TFavSemaphore.Destroy

procedure TFavSemaphore.Release;
{$ifdef Windows}
var MyRV: LongBool;
{$endif}
{$ifdef Linux}
var MyRV: Longint;
{$endif}
begin
  {$ifdef Windows}
  MyRv := ReleaseSemaphore(FSemaphore, 1, nil);
  if MyRv = False then begin
    raise Exception.Create(GetError('RELEASE', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRV := sem_post(@FSemaphore);
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('SEM_POST', fpgeterrno));
    end;
  end
  else begin
    MyRV := sem_post(PSemaphore);
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('SEM_POST', fpgeterrno));
    end;
  end;
  {$endif}
end; // procedure TFavSemaphore.Release

procedure TFavSemaphore.Release(HowMuch: Integer);
{$ifdef Windows}
var MyRV: LongBool;
{$endif}
{$ifdef Linux}
var MyRV: Longint;
    F: Integer;
{$endif}
begin
  {$ifdef Windows}
  MyRv := ReleaseSemaphore(FSemaphore, HowMuch, nil);
  if MyRv = False then begin
    raise Exception.Create(GetError('RELEASE', GetLastError));
  end;
  {$endif}
  {$ifdef Linux}
  // be carefull for linux - this probably need mutex upon this for loop
  if _Name = '' then begin
    for F := 0 to HowMuch-1 do begin
      MyRV := sem_post(@FSemaphore);
      if MyRV <> 0 then begin
        raise Exception.Create(GetError('SEM_POST', fpgeterrno));
      end;
    end;
  end
  else begin
    for F := 0 to HowMuch-1 do begin
      MyRV := sem_post(PSemaphore);
      if MyRV <> 0 then begin
        raise Exception.Create(GetError('SEM_POST', fpgeterrno));
      end;
    end;
  end;
  {$endif}
end;

procedure TFavSemaphore.Wait;
var MyRV: LongInt;
begin
  {$ifdef Windows}
  MyRv := WaitForSingleObject(FSemaphore, Infinite);
  if MyRV <> WAIT_OBJECT_0 then begin
    raise Exception.Create(GetError('WAIT', MyRv));
  end;
  {$endif}
  {$ifdef Linux}
  if _Name = '' then begin
    MyRV := sem_wait(@FSemaphore);
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('SEM_WAIT', fpgeterrno));
    end;
  end
  else begin
    MyRV :=  sem_wait(PSemaphore);
    if MyRV <> 0 then begin
      raise Exception.Create(GetError('SEM_WAIT', fpgeterrno));
    end;
  end;
  {$endif}
end; // procedure TFavSemaphore.Wait

function TFavSemaphore.GetError(InAction: String; InError: Int64): String;
var MyResult: String;
begin
  MyResult := '';
  _LastErroNo:=InError;
  {$ifdef Linux}
  if InAction = 'SEM_OPEN' then begin
    if InError = EACCES then
      MyResult := 'The semaphore exists, but the caller does not have permission to open it. Check /dev/shm directory'
    else if InError = EEXIST then
      MyResult := 'Both O_CREAT and O_EXCL were specified in oflag, but a semaphore with this name already exists.'
    else if InError = EINVAL then
      MyResult := 'value was greater than SEM_VALUE_MAX. name consists of just "/", followed by no other characters. '
    else if InError = EMFILE then
      MyResult := 'The process already has the maximum number of files and open.'
    else if InError = ENAMETOOLONG then
      MyResult := 'name was too long.'
    else if InError = ENFILE then
      MyResult := 'The system limit on the total number of open files has been reached.'
    else if InError = ENOENT then
      MyResult := 'The O_CREAT flag was not specified in oflag and no semaphore with this name exists; or, O_CREAT was specified, but name wasn''t well formed. name consists of just "/", followed by no other characters. '
    else if InError = ENOMEM then
      MyResult := 'Insufficient memory.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'SEM_CLOSE' then begin
    if InError = EINVAL then
      MyResult := 'sem is not a valid semaphore.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'SEM_WAIT' then begin
    if InError = EINTR then
      MyResult := 'The call was interrupted by a signal handler; see signal(7).'
    else if InError = EINVAL then
      MyResult := 'sem is not a valid semaphore. The value of abs_timeout.tv_nsecs is less than 0, or greater than or equal to 1000 million.'
    else if InError = EAGAIN then
      MyResult := 'The operation could not be performed without blocking (i.e., the semaphore currently has the value zero).'
    else if InError = EINTR then
      MyResult := 'The call was interrupted by a signal handler; see signal(7). '
    else if InError = EINTR then
      MyResult := 'The call was interrupted by a signal handler; see signal(7). '
    else if InError = ETIMEDOUT then
      MyResult := 'The call timed out before the semaphore could be locked.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'SEM_POST' then begin
    if InError = EINVAL then
      MyResult := 'sem is not a valid semaphore.'
    else if InError = EOVERFLOW then
      MyResult := 'The maximum allowable value for a semaphore would be exceeded.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'SEM_UNLINK' then begin
    if InError = EACCES then
      MyResult := 'The caller does not have permission to unlink this semaphore.'
    else if InError = ENAMETOOLONG then
      MyResult := 'name was too long. '
    else if InError = ENOENT then
      MyResult := 'There is no semaphore with the given name.'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'SEM_INIT' then begin
    if InError = EINVAL then
      MyResult := 'value exceeds SEM_VALUE_MAX. '
    else if InError = ENOSYS then
      MyResult := 'pshared is nonzero, but the system does not support process-shared semaphores (see sem_overview(7)).'
    else
      MyResult := 'Unknown error';
  end
  else if InAction = 'SEM_DESTROY' then begin
    if InError = EINVAL then
      MyResult := 'sem is not a valid semaphore.'
    else
      MyResult := 'Unknown error';
  end;
  {$endif}
  {$ifdef Windows}
  if (InAction = 'WAIT') or (InAction = 'RELEASE') then begin
    if InError = WAIT_ABANDONED then
      MyResult := 'The specified object is a mutex object that was not released by the thread that owned the mutex object before the owning thread terminated. Ownership of the mutex object is granted to the calling thread and the mutex state is set to nonsignaled. If the mutex was protecting persistent state information, you should check it for consistency.'
    else if InError = WAIT_OBJECT_0 then
      MyResult := 'The state of the specified object is signaled.'
    else if InError = WAIT_TIMEOUT then
      MyResult := 'The time-out interval elapsed, and the object''s state is nonsignaled.'
    else if InError = WAIT_FAILED then
      MyResult := 'The function has failed. To get extended error information, call GetLastError: ' + IntToStr(GetLastError)
    else
      MyResult := 'Unknown error';
  end
  else if (InAction = 'CREATE') or (InAction = 'DESTROY') then begin
    MyResult := 'The function has failed. To get extended error information, call GetLastError: ' + IntToStr(GetLastError);
  end;
  {$endif}

  MyResult := 'Semaphore: ' + _Name + '; Action: ' + InAction + '; Rv: ' + IntToStr(InError) + '; ' + MyResult;
  _LastErrorMsg:=MyResult;
  Result := MyResult;
end;

(*  Example
var Sem: TFavSemaphore;
begin
  // Creating the semaphore.
  Sem := TFavSemaphore.Create;
  .........
  // Requesting access...
  Sem.Wait;
  try
    // Handle the resource...
  finally
    // Giving back the access...
    Sem.Release;
  end; // try-finally
  .........
  // Freeing the semaphore.
  Sem.Free;
end.
*)

initialization
  FAV_THREAD_SYSOBJECTS_CLEAN_TIMEOUT := 10; // One hour (60 * 60)
  FAV_RAISE_SYNCOBJECTS_REGISTRATION_ERROR := False;

  FAV_SYNCOBJECT_STATUS := nil; // If nil program is not in DEBUG MODE
  FAV_THREAD_SYNCOBJECT := nil;
  {$IFOPT D-}
    // this switch check is it compile for DEBUG - this can be use to create thread sync objects
    FAV_SYNCOBJECT_STATUS := nil; // If nil program is not in DEBUG MODE
    FAV_THREAD_SYNCOBJECT := nil;
    FAV_SYNCOBJECT_STATUS := TFavSyncObjectsStatuses.Create;
    FAV_THREAD_SYNCOBJECT := TThreadSyncObjectStatus.Create(False);  // start immediatelly  just for test
  {$ELSE}
    FAV_SYNCOBJECT_STATUS := TFavSyncObjectsStatuses.Create;
    FAV_THREAD_SYNCOBJECT := TThreadSyncObjectStatus.Create(False);  // start immediatelly
  {$ENDIF}



finalization
  if FAV_THREAD_SYNCOBJECT <> nil then begin
    FreeAndNil(FAV_THREAD_SYNCOBJECT);
  end;
  if FAV_SYNCOBJECT_STATUS <> nil then begin
    FreeAndNil(FAV_SYNCOBJECT_STATUS);
  end;

end.

