unit ThreadTest;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FavSemaphore, DateUtils;

type
  { TThreadTestWriteReadSemaphore }

  TThreadTestWriteReadSemaphore = class(TThread)
  private

  protected
    procedure Execute; override;
  public
    _PSemaphore: TFavWriterReaderSemaphoreWriterPriority;
    _AMIReader: Boolean;
    _BShutdown, _BTerminated: Boolean;
    _Counter: Int64;
    _ThreadID: Int64;
    _Value: Int64;
    constructor Create(InThreadID: Int64;
                       InSemaphore: TFavWriterReaderSemaphoreWriterPriority;
                       InAmIReader: Boolean;
                       CreateSuspended: Boolean;
                       const StackSize: SizeUInt = DefaultStackSize);
    destructor Destroy; override;
    procedure ShutDownThread;
  end;

procedure FavLogToConsole(InString: String);

const
  CONST_READER_COUNT = 16; // NUMBER OF READER THREADS
  CONST_WRITER_COUNT = 2; // NUMBER OF WRITER THREAD

  CONST_READER_SLEEP = 1; // LENGTH IN MS OF READER JOB
  CONST_WRITER_SLEEP = 10; // LENGTH IN MS OF WRITER JOB


var MY_GLOBAL_VALUE: Int64;
    MY_GLOBAL_SEMAPHORE: TFavWriterReaderSemaphoreWriterPriority;
    MY_READERS: array[0..CONST_READER_COUNT] of TThreadTestWriteReadSemaphore;
    MY_WRITERS: array[0..CONST_WRITER_COUNT] of TThreadTestWriteReadSemaphore;

implementation

procedure FavSaveToFile(InFileName: String; const InString: String;
  InRaiseError: Boolean);
var MyFile: Text;
    MyFileAssigned: Boolean;
    MyFileOpened: Boolean;
begin
  MyFileAssigned:=False;
  MyFileOpened:=False;
  try
    AssignFile(MyFile, InFileName);
    MyFileAssigned:=True;
    Rewrite(MyFile);
    MyFileOpened:=True;
    Write(MyFile, InString);
    Flush(MyFile);
    CloseFile(MyFile);
    MyFileOpened:=False;
    MyFileAssigned:=False;
  except
    on E: Exception do begin
      if MyFileAssigned then begin
        try
          if MyFileOpened then begin
            CloseFile(MyFile);
          end;
        except
        end;
      end;
      if InRaiseError then begin;
        raise;
      end;
    end;
  end;
end;

procedure FavLogToConsole(InString: String);
begin
    if IsConsole then begin
    try
      {$ifdef Windows}
      Writeln(StdOut, FormatDateTime('yyyy-mm-dd hh:nn:ss.zzz', now), ': ', InString);
      {$else}
      Write(StdOut, FavToNiceStringDateAndTimeAndMSec(FavUTCNow), ': ', InString, #13#10);
      {$endif}
    except
    end;
  end;
end;


{ TThreadTestWriteReadSemaphore }

procedure TThreadTestWriteReadSemaphore.Execute;
var MyUntil: TDateTime;
begin
  _Counter := 0;
  _Value := 0;
  { Write your thread code here }
  _BTerminated := False;
  while true do begin
    if _BShutdown then begin
      Break;
    end;

    try
      (*
      if (_Counter mod 100000) = 0 then begin
        if _AMIReader then begin
          FavLogToConsole('Reader: ' + FavToString(_AMIReader) + ', with ID: ' + FavToString(_ThreadID) + ', have value: ' + FavToString(_Value) + ' and counter is ' + FavToString(_Counter));
        end
        else begin
          FavLogToConsole('Writer: ' + FavToString(_AMIReader) + ', with ID: ' + FavToString(_ThreadID) + ', have value: ' + FavToString(_Value) + ' and counter is ' + FavToString(_Counter));
        end;
      end;
      *)

      //sleep(1);


      if _AMIReader then begin
        _PSemaphore.WaitForRead(IntToStr(_ThreadID));
        _Counter := _Counter + 1;
        if CONST_READER_SLEEP > 0 then begin
          MyUntil := Now + CONST_READER_SLEEP * OneMillisecond;
          while Now < MyUntil do begin

          end;
          // Sleep(CONST_READER_SLEEP);
        end;
        _Value := MY_GLOBAL_VALUE;
        _PSemaphore.ReleaseRead(IntToStr(_ThreadID));
      end
      else begin
        //FavLogToConsole('Writer: ' + ', with ID: ' + FavToString(_ThreadID) + ' is waiting');
        _PSemaphore.WaitForWrite(IntToStr(_ThreadID));
        _Counter := _Counter + 1;
        //FavLogToConsole('Writer: ' + ', with ID: ' + FavToString(_ThreadID) + ' got semaphore');
        if CONST_WRITER_SLEEP > 0 then begin
          MyUntil := Now + CONST_WRITER_SLEEP * OneMillisecond;
          while Now < MyUntil do begin

          end;
          // Sleep(CONST_WRITER_SLEEP);  // simulate write job length (100 times bigger then read)
        end;
        MY_GLOBAL_VALUE := MY_GLOBAL_VALUE + 1;
        _Value := MY_GLOBAL_VALUE;
        //FavLogToConsole('Writer: ' + ', with ID: ' + FavToString(_ThreadID) + ' will release semaphore, and set value: ' + FavToString(MY_GLOBAL_VALUE));
        _PSemaphore.ReleaseWrite(IntToStr(_ThreadID));
        // _PSemaphore.ReleaseWrite(FavToString(_ThreadID));
        //FavLogToConsole('Writer: ' + ', with ID: ' + FavToString(_ThreadID) + ' released semaphore');
      end;

    except
      FavSaveToFile('test.file', 'Progam exit', false);
    end;

  end;
  _BTerminated := True;
end;

constructor TThreadTestWriteReadSemaphore.Create(InThreadID: Int64;
  InSemaphore: TFavWriterReaderSemaphoreWriterPriority; InAmIReader: Boolean;
  CreateSuspended: Boolean; const StackSize: SizeUInt);
begin
  inherited Create(CreateSuspended, StackSize);
  _ThreadID := InThreadID;
  _AMIReader := InAmIReader;
  _PSemaphore := InSemaphore;
  _BShutdown := False;
  _BTerminated := True;
end;

destructor TThreadTestWriteReadSemaphore.Destroy;
begin
  ShutDownThread;
  FavLogToConsole('END Thread Reader: ' + BoolToStr(_AMIReader) + ', with ID: ' + IntToStr(_ThreadID) + ', have value: ' + IntToStr(_Value) + ' and counter is ' + IntToStr(_Counter));
  inherited Destroy;
end;

procedure TThreadTestWriteReadSemaphore.ShutDownThread;
begin
  while _BTerminated = False do begin
    _BShutdown := true;
    Sleep(2);
  end;
end;

end.

